import logging
import os
import platform
from pathlib import Path
from typing import Optional, cast

import keyring
import pylast
import spotipy
from decouple import AutoConfig
from spotipy.cache_handler import CacheFileHandler
from spotipy.oauth2 import SpotifyOAuth

from gather_scrobble.utils import get_logger

config = AutoConfig(os.getcwd())  # search for settings in the current dir
USE_CRYPTFILE = config("USE_CRYPTFILE", cast=bool, default=False)
CONFIG_FOLDER = "gather-scrobble"

logger = get_logger(
    "gather-scrobble", logging.DEBUG if USE_CRYPTFILE else logging.INFO
)


class Credentials:
    def __init__(
        self,
        gather_api_key: str,
        lastfm_api_key: Optional[str],
        lastfm_api_secret: Optional[str],
        lastfm_username: Optional[str],
        spotify_client_id: Optional[str],
        spotify_client_secret: Optional[str],
        spotify_client_redirect_uri: Optional[str],
    ) -> None:
        self.gather_api_key = gather_api_key
        self.lastfm_api_key = lastfm_api_key
        self.lastfm_api_secret = lastfm_api_secret
        self.lastfm_username = lastfm_username
        self.spotify_client_id = spotify_client_id
        self.spotify_client_secret = spotify_client_secret
        self.spotify_client_redirect_uri = spotify_client_redirect_uri

    @property
    def has_lastfm(self) -> bool:
        return (
            self.lastfm_api_key is not None
            and self.lastfm_api_secret is not None
            and self.lastfm_username is not None
        )

    @property
    def has_spotify(self) -> bool:
        return (
            self.spotify_client_id is not None
            and self.spotify_client_secret is not None
            and self.spotify_client_redirect_uri is not None
        )

    @property
    def is_valid(self) -> bool:
        return self.has_lastfm or self.has_spotify


def _get_config_value(var_name: str) -> Optional[str]:
    keyring_value = keyring.get_password("gather-scrobble", var_name)
    env_var = config(var_name, default=None)
    return keyring_value or cast(Optional[str], env_var)


def get_credentials():
    if USE_CRYPTFILE:
        from keyrings.cryptfile.cryptfile import CryptFileKeyring

        kr = CryptFileKeyring()
        kr.keyring_key = config("KEYRING_CRYPTFILE_PASSWORD", default="secret")
        keyring.set_keyring(kr)
    gather_api_key = _get_config_value("GATHER_API_KEY")
    if gather_api_key is None or str(gather_api_key).strip() == "":
        raise Exception("Missing 'GATHER_API_KEY' value")
    lastfm_api_key = _get_config_value("LASTFM_API_KEY")
    lastfm_api_secret = _get_config_value("LASTFM_API_SECRET")
    lastfm_username = _get_config_value("LASTFM_USERNAME")
    spotify_client_id = _get_config_value("SPOTIFY_CLIENT_ID")
    spotify_client_secret = _get_config_value("SPOTIFY_CLIENT_SECRET")
    spotify_client_redirect_uri = _get_config_value(
        "SPOTIFY_CLIENT_REDIRECT_URI"
    )
    return Credentials(
        gather_api_key,
        lastfm_api_key,
        lastfm_api_secret,
        lastfm_username,
        spotify_client_id,
        spotify_client_secret,
        spotify_client_redirect_uri,
    )


def get_config_folder_path() -> Path:
    if "windows" in platform.system().lower():
        config_dir = Path(os.environ.get("APPDATA", "config"), CONFIG_FOLDER)
    else:
        config_dir = Path(
            os.environ.get("XDG_CONFIG_HOME", Path("~/.config").expanduser()),
            CONFIG_FOLDER,
        )
    if not config_dir.exists():
        config_dir.mkdir(parents=True, exist_ok=True)
    return config_dir


def get_lastfm_session_key(network: pylast.LastFMNetwork) -> str:
    session_key_file = Path(
        get_config_folder_path(), f"lastfm_user_{network.username}"
    )
    if not session_key_file.exists():
        skg = pylast.SessionKeyGenerator(network)
        url = skg.get_web_auth_url()

        print(f"Please authorize this script to access your account: {url}\n")
        import time
        import webbrowser

        webbrowser.open(url)

        while True:
            try:
                session_key = skg.get_web_auth_session_key(url)
                with open(session_key_file, "w") as f:
                    f.write(session_key)
                break
            except pylast.WSError:
                time.sleep(1)
    return session_key_file.read_text()


def get_lastfm_client(credentials: Credentials):
    if not credentials.has_lastfm:
        raise Exception("last.fm not configured.")
    network = pylast.LastFMNetwork(
        cast(str, credentials.lastfm_api_key),
        cast(str, credentials.lastfm_api_secret),
    )
    network.username = credentials.lastfm_username
    network.session_key = get_lastfm_session_key(network)
    return network.get_user(credentials.lastfm_username)


class CustomCacheHandler(CacheFileHandler):
    def __init__(self, cache_path=None, username=None):
        custom_cache_path = Path(get_config_folder_path(), "spotify.cache")
        super().__init__(custom_cache_path, username)


def get_spotify_client(credentials: Credentials):
    if not credentials.has_spotify:
        raise Exception("Spotify not configured.")

    auth_manager = SpotifyOAuth(
        credentials.spotify_client_id,
        credentials.spotify_client_secret,
        credentials.spotify_client_redirect_uri,
        scope=["user-read-currently-playing", "user-read-playback-state"],
        cache_handler=CustomCacheHandler(),
        open_browser=not USE_CRYPTFILE,
    )
    return spotipy.Spotify(auth_manager=auth_manager)



import asyncio
import logging
import queue
import random
import threading
from itertools import cycle
from sys import stdout as terminal
from time import sleep
from typing import Optional

from gather_client_ws import GatherClient

from gather_scrobble.config import (
    Credentials,
    get_credentials,
    get_lastfm_client,
    get_spotify_client,
    logger,
)

BARS = "𝄖𝄗𝄘𝄙𝄚𝄛"

PLAYING_ANIMATION = [
    [1, 2, 0, 1, 4],
    [2, 2, 0, 2, 4],
    [3, 2, 1, 3, 4],
    [4, 3, 1, 4, 4],
    [5, 3, 2, 5, 3],
    [0, 3, 2, 0, 3],
    [1, 4, 3, 1, 3],
    [2, 4, 3, 2, 3],
    [3, 4, 4, 3, 2],
    [4, 5, 4, 4, 2],
    [5, 5, 5, 5, 2],
    [0, 5, 5, 0, 2],
]

EMOJIS = ["🎼", "🎵", "🎶", "🎧", "📻", "🎷", "🪗", "🎸", "🎹", "🎺", "🎻", "🪕", "🥁", "🪘"]


def parse_source(credentials: Credentials, source: str) -> str:
    if source == "any":
        if credentials.has_lastfm:
            return "lastfm"
        return "spotify"
    return source


class ScrobbleClient:
    def __init__(
        self, credentials: Credentials, source: str, emojis: str
    ) -> None:
        self.credentials = credentials
        self.source = parse_source(credentials, source)
        self.emojis = [e for e in emojis] if emojis != "" else [""]
        self._lastfm = None
        self._spotify = None
        if self.source == "lastfm":
            self._lastfm = get_lastfm_client(self.credentials)
        else:
            self._spotify = get_spotify_client(self.credentials)

    def get_lastfm_now_playing(self) -> Optional[str]:
        if self._lastfm is None:
            return None
        track = self._lastfm.get_now_playing()
        if track is None:
            return None
        artist_name = (
            track.artist.get_name() if track.artist is not None else ""
        )
        return f"{track.get_title()} - {artist_name}"

    def get_spotify_now_playing(self) -> Optional[str]:
        if self._spotify is None:
            return None
        track = self._spotify.current_user_playing_track()
        if track is None:
            return None
        name = track.get("item", {}).get("name", "")
        artists = track.get("item", {}).get("artists", [])
        artist_name = artists[0].get("name") if len(artists) > 0 else ""
        if name.strip() == "" or artist_name.strip() == "":
            return None
        return f"{name} - {artist_name}"

    def get_now_playing(self) -> Optional[str]:
        if self.source == "lastfm":
            return self.get_lastfm_now_playing()
        return self.get_spotify_now_playing()


async def clear_status(client):
    await client.set_emoji_status("")
    await client.set_text_status("")


async def set_status(client, emoji: str, text: str):
    await client.set_emoji_status(emoji)
    await client.set_text_status(text)
    logger.debug("%s - %s", emoji, text)


async def scrobble(client, scrobble_client):
    q = queue.Queue()

    def show_now_playing():
        default_msg = "Nothing is playing now 😕"
        msg = ""
        for bars in cycle(PLAYING_ANIMATION):
            if q.qsize() > 0:
                msg = q.get()
            if msg.strip() == "":
                to_write = default_msg
            else:
                to_write = f"{msg} {''.join(BARS[index] for index in bars)}"
            # No show the playing now when running in docker
            # to avoid messing with the logs.
            if logger.level > logging.DEBUG:
                terminal.write(f"\033[2K\r{to_write}")
                terminal.flush()
            sleep(0.1)

    current_now_playing = None
    threading.Thread(target=show_now_playing, daemon=True).start()
    retry = 0
    logger.debug("application started")
    while True:
        try:
            now_playing = scrobble_client.get_now_playing()
            if now_playing is None:
                if current_now_playing is not None:
                    await clear_status(client)
                current_now_playing = None
            elif current_now_playing != now_playing:
                emoji = random.choice(scrobble_client.emojis)
                current_now_playing = now_playing
                await set_status(client, emoji, now_playing)
                if emoji:
                    q.put(f"{emoji} - {current_now_playing}")
                else:
                    q.put(f"{current_now_playing}")
            await asyncio.sleep(15)
            retry = 0
        except KeyboardInterrupt:
            break
        except Exception:
            # The retry will ensure the problem is not temporary
            retry += 1
            if retry >= 3:
                break
            await asyncio.sleep(2)


async def start(space_id, source, emojis):
    logger.debug("loading credentials")
    credentials = get_credentials()
    if not credentials.is_valid:
        raise Exception("At least one source needs to be configured.")
    scrobble_client = ScrobbleClient(credentials, source, emojis)
    client = GatherClient(
        credentials.gather_api_key, space_id, log_level=logging.ERROR
    )
    logger.debug("starting application")
    await client.run(scrobble, scrobble_client)



import asyncio
import logging

from gather_client_ws import GatherClient
from tabulate import tabulate

from gather_scrobble.config import (
    get_credentials,
    get_lastfm_client,
    get_spotify_client,
)


async def gather_heartbeat(client):
    await client.client_heartbeat()


async def test_gather_connection(credentials, space_id):
    client = GatherClient(
        credentials.gather_api_key, space_id, log_level=logging.ERROR
    )
    await client.run(gather_heartbeat)


def test_configuration(space_id):
    credentials = get_credentials()
    result = []
    print("Testing connection with Gather...")
    try:
        asyncio.get_event_loop().run_until_complete(
            test_gather_connection(credentials, space_id)
        )
        result.append(["Gather", True])
        print("Success")
    except Exception:
        result.append(["Gather", False])
        print("Failed")
    print("Testing connection with last.fm...")
    try:
        lastfm = get_lastfm_client(credentials)
        lastfm.get_now_playing()
        result.append(["last.fm", True])
        print("Success")
    except Exception:
        result.append(["last.fm", False])
        print("Failed")

    print("Testing connection with Spotify...")
    try:
        spotify = get_spotify_client(credentials)
        spotify.current_user_playing_track()
        result.append(["Spotify", True])
        print("Success")
    except Exception:
        result.append(["Spotify", False])
        print("Failed")

    print()
    print(tabulate(result, ["Service", "Working?"], tablefmt="heavy_grid"))



import asyncio
from contextlib import suppress

from docopt import docopt

from gather_scrobble.config import logger
from gather_scrobble.info import show_info
from gather_scrobble.scrobble import start
from gather_scrobble.test import test_configuration


def main():
    """Gather Scrobble v0.0.7
    Usage:
        gather-scrobble start <space_id> [--source SOURCE] [--emojis EMOJIS]
        gather-scrobble info
        gather-scrobble test <space_id>

    Arguments:
        <space_id>          Gather space id.

    Options:
        -h --help           Show this screen.
        -s --source SOURCE  Scrobble source, if not defined the tool will try
                            all available sources configured.
                            Can be: lastfm, spotify, or any.
                            If 'any' it will use the first source configured in
                            the priority: lastfm -> spotify [default: any]
        -e --emojis EMOJIS  It is possible to customize the emojis, by setting
                            the list of emojis that will be chosen randomly,
                            and also you don't like emojis, you can set an
                            empty string here. [default: 🎼🎵🎶🎧📻🎷🎸🎹]

    """
    arguments = docopt(main.__doc__ or "", version="Gather Scrobble 0.0.7")
    if arguments["start"]:
        if arguments["--source"] not in ["lastfm", "spotify", "any"]:
            raise Exception(f"Invalid source: {arguments['--source']}")
        with suppress(KeyboardInterrupt):
            asyncio.get_event_loop().run_until_complete(
                start(
                    arguments["<space_id>"],
                    arguments["--source"],
                    arguments["--emojis"],
                )
            )
        logger.debug("bye")
        print("bye")
    if arguments["info"]:
        show_info()
    if arguments["test"]:
        test_configuration(arguments["<space_id>"])



from tabulate import tabulate

from gather_scrobble.config import get_credentials


def show_info():
    credentials = get_credentials()
    print("Gather Scrobble")
    print(
        tabulate(
            [
                ["Gather", True],
                ["last.fm", credentials.has_lastfm],
                ["Spotify", credentials.has_spotify],
            ],
            ["Service", "Configured"],
            tablefmt="heavy_grid",
        )
    )



#!/usr/bin/env python
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import os
import re
import sys

try:
    import pypandoc

    readme = pypandoc.convert_file("README.md", "rst")
except (IOError, ImportError):
    readme = ""


package = "gather_scrobble"
requirements = [
    "gather-client-ws<1",
    "keyring==23.13.1",
    "keyrings.cryptfile==1.3.9",
    "python-decouple==3.8",
    "tabulate==0.9.0",
    "pylast==5.1.0",
    "spotipy==2.22.1",
    "docopt==0.6.2",
]
test_requirements = []


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, "__init__.py")).read()
    return re.search(
        "^__version__ = ['\"]([^'\"]+)['\"]", init_py, re.MULTILINE
    ).group(1)


def get_author(package):
    """
    Return package author as listed in `__author__` in `init.py`.
    """
    init_py = open(os.path.join(package, "__init__.py")).read()
    return re.search(
        "^__author__ = ['\"]([^'\"]+)['\"]", init_py, re.MULTILINE
    ).group(1)


def get_email(package):
    """
    Return package email as listed in `__email__` in `init.py`.
    """
    init_py = open(os.path.join(package, "__init__.py")).read()
    return re.search(
        "^__email__ = ['\"]([^'\"]+)['\"]", init_py, re.MULTILINE
    ).group(1)


# python setup.py register
if sys.argv[-1] == "publish":
    os.system("python setup.py sdist upload")
    args = {"version": get_version(package)}
    print("You probably want to also tag the version now:")
    print("  git tag -a %(version)s -m 'version %(version)s'" % args)
    print("  git push --tags")
    sys.exit()


setup(
    name="gather-scrobble",
    version=get_version(package),
    description="Gather Scrobble - Scrobble your last.fm or Spotify activity to the Gather status.",
    long_description=readme,
    author=get_author(package),
    author_email=get_email(package),
    url="https://github.com/pyanderson/gather-scrobble",
    packages=[
        "gather_scrobble",
    ],
    package_dir={"gather_scrobble": "gather_scrobble"},
    include_package_data=True,
    install_requires=requirements,
    entry_points={
        "console_scripts": ["gather-scrobble = gather_scrobble.cli:main"]
    },
    license="MIT",
    zip_safe=False,
    keywords=["gather", "gather-town", "lastfm", "spotify", "scrobble"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    test_suite="tests",
    tests_require=test_requirements,
)



import logging


def get_logger(name: str, level: int = logging.INFO):
    logger = logging.getLogger(name)
    logger.setLevel(level)
    ch = logging.StreamHandler()
    ch.setLevel(level)
    formatter = logging.Formatter("[%(asctime)s][%(levelname)s] %(message)s")
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger
